#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________
#

from pyutilib.math.numtypes import infinity, nan, is_nan, is_finite
from pyutilib.math.util import approx_equal, as_number, isint, argmax, argmin, mean, median, factorial, perm, isclose
